
// FileNotes.java
// Andrew Davison, March 2009, ad@fivedots.coe.psu.ac.th

/* Load a file line-by-line into an ArrayList, analyze each line
   of code, and convert those lines containing a keyword into 
   musical notes. Those lines without keywords are stored as ""
   in the list.

   The getNote() method returns the musical note (as a string)
   for a specified line (or null).


*/

import java.io.*;
import java.util.*;


public class FileNotes
{
  // the keywords are mapped to notes using two arrays

  /* for a complete list of keywords see 
     http://java.sun.com/docs/books/tutorial/java/nutsandbolts/_keywords.html  */
  private final String[] keywords = 
           { "else", "break", "switch", "case", "finally", "catch", "for", 
             "if", "try", "continue", "return", "default", "do", "while"
           };

   // C D E F G A B  in the 4th and 5th octave on a piano 
  private final String[] notes = 
           { "C4", "D4", "E4", "F4", "G4", "A4", "B4",
             "C5", "D5", "E5", "F5", "G5", "A5", "B5"
           };


  private ArrayList<String> codeNotes;    
     // each line of the input file is represented by a musical note string, or ""


  public FileNotes(String fileName)
  {
    codeNotes = new ArrayList<String>();
    
    int lineNum = 0;
    String line = null;
    BufferedReader in = null;
    String[] words;

    System.out.println("Analyzing " + fileName);
    try {
      in = new BufferedReader(new FileReader(fileName));
      while ((line = in.readLine()) != null) {
        lineNum++;
        words = line.split("\\W+");    // split each line on word boundaries
        storeNote(words, lineNum);
      }
    }
    catch (IOException ex) {
      System.out.println("Problem reading " + fileName);
    } 
    finally {
      try {
        if (in != null)
          in.close();
      }
      catch (IOException e) {}
    }
  }  // end of showLines()



  private void storeNote(String[] words, int lineNum)
  /* If a keyword is found in words, then store the corresponding
     note in the codeNotes list. 
     If no keyword is found, then store "".
     We stop searching as soon as one keyword is found, so a
     line containing multiple keywords will not be represented by multiple
     notes in the list. */
  {
     for(String word : words)
       for (int i=0; i < keywords.length; i++)
         if (word.equalsIgnoreCase(keywords[i])) {
           codeNotes.add(notes[i]);
           System.out.println("  " + lineNum + ": found " + keywords[i]);
           return;
         }
       
     codeNotes.add("");      // no keyword found
  }  // end of storeNote()



  public String getNote(int lineNum)
  // lineNum starts at 1, not 0
  // return a musical note sring or null (not "")
  {
    if (codeNotes == null) {
      System.out.println("No code to show");
      return null;
    }

    if ((lineNum < 1) || (lineNum > codeNotes.size())) {
      System.out.println("Line no. out of range");
      return null;
    }

    String note = codeNotes.get(lineNum-1); 
    if (note.equals("")) {
      // System.out.println("No note for line " + lineNum);
      return null;
    }

    return note;
  }  // end of getNote()

    
/*
  // -------------------- test rig ----------------------

  public static void main(String[] args)
  {  
     FileNotes sl = new FileNotes("FileNotes.java"); 
     System.out.println( sl.getNote(56));
     System.out.println( sl.getNote(1));
     System.out.println( sl.getNote(2000));
  }
*/

}  // end of FileNotes class
