
// NotesPlayer.java
// Andrew Davison, April 2009, ad@fivedots.coe.psu.ac.th

/*  Play a musical note for a specified amount of time 
    using a MIDI synthesizer. The note is supplied as
    a string (e.g. "C4", "B5#") and the duration is in ms.
*/


import javax.sound.midi.*;
import java.util.*;


public class NotesPlayer
{

  // channel used by the synthesizer
  private static final int CHANNEL = 0;

  private static final int VOLUME = 90;  // fixed volume for notes

  /* The note offsets use the "C" major scale, which
     has the order "C D E F G A B", but the offsets are
     stored in the order "A B C D E F G" to simplify their 
     lookup. */
  private static final int[] cOffsets =  {9, 11, 0, 2, 4, 5, 7};
                                        // A  B  C  D  E  F  G
  private static final int C4_KEY = 60;    
        // C4 is the "C" in the 4th octave on a piano 
  private static final int OCTAVE = 12;    // note size of an octave


  private Receiver receiver;



  public NotesPlayer()
  { 
    receiver = findReceiver("Java Sound");  // or perhaps "Microsoft GS"
  } // end of NotesPlayer()



  private Receiver findReceiver(String name)
  // find the receiver MIDI device called name
  {
    System.out.println("Searching for device: " + name);
    try {
      MidiDevice.Info[] devices = MidiSystem.getMidiDeviceInfo();
      for(MidiDevice.Info devInfo : devices)
        if (devInfo.getName().startsWith(name)) {
          MidiDevice dev = MidiSystem.getMidiDevice(devInfo);
          if (dev.getMaxReceivers() != 0) {
            System.out.println("Found: " + devInfo.getDescription());
            dev.open();
            System.out.println("Opening the Receiver");
            return dev.getReceiver();
          }
        }
      System.out.println("Device not found");
    }
    catch (MidiUnavailableException e)
    {  System.out.println(e);  }

    return null;
  }  // end of findReceiver()



  public void play(String noteStr, int duration) 
  // play note for specified duration
  {
    if (receiver == null) {
      System.out.println("No synthesizer to play note: " + noteStr);
      return;
    }

    System.out.println("Playing the note " + noteStr + " for " + duration);

    int note = getKey(noteStr);

    ShortMessage message = new ShortMessage();
    try {
      message.setMessage(ShortMessage.NOTE_ON, CHANNEL, note, VOLUME);
      receiver.send(message, -1);

      wait(duration); // in ms

      message.setMessage(ShortMessage.NOTE_OFF, CHANNEL, note, VOLUME);
      receiver.send(message, -1);

      // System.out.println("Played a note");
    }
    catch (InvalidMidiDataException e) 
    {  System.out.println(e);  }
  }  // end of play()



  private int getKey(String noteStr)
  /* Convert a note string (e.g. "C4", "B5#") into a key. */
  {
    char[] letters = noteStr.toCharArray();

    if (letters.length < 2) {
      System.out.println("Incorrect note syntax; using C4");
      return C4_KEY;
    }

    // look at note letter in letters[0]
    int c_offset = 0;
    if ((letters[0] >= 'A') && (letters[0] <= 'G'))
      c_offset = cOffsets[letters[0] - 'A'];
    else
      System.out.println("Incorrect letter: " + letters[0] + ", using C");

    // look at octave number in letters[1]
    int range = C4_KEY;
    if ((letters[1] >= '0') && (letters[1] <= '9'))
      range = OCTAVE * (letters[1] - '0' + 1);  // plus 1 for midi
    else
      System.out.println("Incorrect number: " + letters[1] + ", using 4");

    // look at optional sharp in letters[2]
    int sharp = 0;
    if ((letters.length > 2) && (letters[2] == '#'))
      sharp = 1;    // a sharp is 1 note higher 
                    // (represented by the black keys on a piano)
    int key = range + c_offset + sharp;
    // System.out.println("note: " + noteStr + "; key: " + key);

    return key;
  }  // end of getKey()


  private void wait(int duration)
  {
    try {
      Thread.sleep(duration); // in ms
    } 
    catch (InterruptedException e) 
    { e.printStackTrace(); }
  }  // end of wait()


  public void closeDown()
  { if (receiver != null)
	  receiver.close();
  }


/*
  // -------------------- test rig ----------------------


  public static void main(String[] args)
  { 
    NotesPlayer player = new NotesPlayer();

    player.play("C4#", 500);
    player.play("D4", 1000);
    player.play("E4", 1000);

	player.closeDown();
  } // end of main()
*/

} // end of NotesPlayer class
