
// SoundEventsMonitor.java

/* Robert Field and Minoru Terada, September 2005
   Iman_S, June 2008  */
// Andrew Davison, ad@fivedots.coe.psu.ac.th, April 2009

/* Monitor incoming JDI events for a program running in the JVM.
   This is a simplified version of JDIEventMonitor.java from 
   SimpleTrace.java.

   When a method is called, a word will be spoken using the free-TTS
   speech synthesizer. When keywords in the code are encountered,
   musical notes will be played. When the program starts, ends, and when
   a method returns, sound clips are played. 

   Sound messages are generated here and sent via a blocking queue to
   a sound generation thread (SoundGen) which converts the messages into
   audio.

   Three types of messages are placed onto the blocking queue going
   to the SoundGen thread:
     $w <word to be said>                   // done in methodEntryEvent()

     $n <note to be played>                 // done in stepEvent()

     $c <wav filename (a clip) to be played>  <duration(ms)>   
                        // done in methodEntryEvent() & methodExitEvent()


  CHANGES from JDIEventMonitor:

    * uses AppNotes, not ShowCode
    * no watching of fields in a class (no setFieldsWatch(), fieldWatchEvent())
    * no printing of locals or fields when stepping through code
    * no threadDeathEvent()
    * no classUnloadEvent()

    * modified stepEvent() sends a musical note message for a line (if there is one)
    * classPrepareEvent() builds appNotes and method words list
    * methodEntryEvent() reports method name
*/


import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;

import java.util.*;
import java.util.concurrent.*;


public class SoundEventsMonitor extends Thread
{
  // exclude events generated for these classes
  private final String[] excludes = { "java.*", "javax.*", "sun.*", "com.sun.*"};

  private final VirtualMachine vm;   // the JVM
  private boolean connected = true;  // connected to VM?
  private boolean vmDied;            // has VM death occurred?


  // musical data objects
  private AppNotes appNotes;    
         // stores musical notes representing the application code
  private MethodWords methodWords;   
            // stores the words that will be spoken when methods are called

  // sound generation objects
  private BlockingQueue<String> sndQueue;   // holds sound messages
  private SoundGen soundGen;     // processes the sound messages



  public SoundEventsMonitor(VirtualMachine jvm)
  {
    super("SoundEventsMonitor");
    vm = jvm;

    appNotes = new AppNotes();
    methodWords = new MethodWords();

    // initialize the message queue and the sound generation thread
    sndQueue = new LinkedBlockingQueue<String>();
    soundGen = new SoundGen(sndQueue);
    soundGen.start();

    setEventRequests();
  }  // end of SoundEventsMonitor()



  private void setEventRequests()
  /* Create and enable the event requests for the events
     we want to monitor in the running program. */
  {
    EventRequestManager mgr = vm.eventRequestManager();

    MethodEntryRequest menr = mgr.createMethodEntryRequest(); // report method entries
    for (int i = 0; i < excludes.length; ++i)
      menr.addClassExclusionFilter(excludes[i]);
    menr.setSuspendPolicy(EventRequest.SUSPEND_EVENT_THREAD);
    menr.enable();
	
    MethodExitRequest mexr = mgr.createMethodExitRequest();   // report method exits
    for (int i = 0; i < excludes.length; ++i)
      mexr.addClassExclusionFilter(excludes[i]);
    mexr.setSuspendPolicy(EventRequest.SUSPEND_EVENT_THREAD);
    mexr.enable();

    ClassPrepareRequest cpr = mgr.createClassPrepareRequest(); // report class loads
    for (int i = 0; i < excludes.length; ++i)
      cpr.addClassExclusionFilter(excludes[i]);
    // cpr.setSuspendPolicy(EventRequest.SUSPEND_ALL);
    cpr.enable();

    ThreadStartRequest tsr = mgr.createThreadStartRequest();  // report thread starts
    tsr.enable();

  }  // end of setEventRequests()



  public void run()
  // process JDI events as they arrive on the event queue
  { 
    EventQueue queue = vm.eventQueue();
    while (connected) {
      try {
        EventSet eventSet = queue.remove();
        for(Event event : eventSet)
          handleEvent(event);
        eventSet.resume();
      }
      catch (InterruptedException e) { }  // Ignore
      catch (VMDisconnectedException discExc) {
        handleDisconnectedException();
        break;
      }
    }
  }  // end of run()



  private void handleEvent(Event event)
  // process a JDI event
  {
    // method events
    if (event instanceof MethodEntryEvent)
      methodEntryEvent((MethodEntryEvent) event);
    else if (event instanceof MethodExitEvent)
      methodExitEvent((MethodExitEvent) event);

    // class events
    else if (event instanceof ClassPrepareEvent)
      classPrepareEvent((ClassPrepareEvent) event);

    // thread events
    else if (event instanceof ThreadStartEvent)
      threadStartEvent((ThreadStartEvent) event);

    // step event -- a line of code is about to be executed
    else if (event instanceof StepEvent)
      stepEvent((StepEvent) event);

    // VM events
    else if (event instanceof VMStartEvent)
      vmStartEvent((VMStartEvent) event);
    else if (event instanceof VMDeathEvent)
      vmDeathEvent((VMDeathEvent) event);
    else if (event instanceof VMDisconnectEvent)
      vmDisconnectEvent((VMDisconnectEvent) event);

    else
      throw new Error("Unexpected event type");
  }  // end of handleEvent()


  private synchronized void handleDisconnectedException()
  /* A VMDisconnectedException has occurred while dealing with
     another event. Flush the event queue, dealing only
     with exit events (VMDeath, VMDisconnect) so that things 
     terminate correctly. */
  {
    EventQueue queue = vm.eventQueue();
    while (connected) {
      try {
        EventSet eventSet = queue.remove();
        for(Event event : eventSet) {
          if (event instanceof VMDeathEvent)
            vmDeathEvent((VMDeathEvent) event);
          else if (event instanceof VMDisconnectEvent)
            vmDisconnectEvent((VMDisconnectEvent) event);
        }
        eventSet.resume(); // resume the VM
      }
      catch (InterruptedException e) { }  // ignore
    }
  }  // end of handleDisconnectedException()



  // -------------------- method event handling  ---------------



  private void methodEntryEvent(MethodEntryEvent event)
  // entered a method but no code executed yet
  // play a clip if the method is main(), and say the method word
  { 
    Method meth = event.method();
    String className = meth.declaringType().name();

    System.out.println();
    if (meth.isConstructor())     // use class name for the constructor
      System.out.println("entered " + className + " constructor");
    else
      System.out.println("entered " + className +  "." + meth.name() +"()");

    if (meth.name().equals("main"))   // program is starting so play a clip
      saySound("$c windChime 1930");  

    // report method entry by saying a word
    String word = methodWords.getWord(className, meth);
    if (word != null) {
      System.out.println("  saying " + word);
      saySound("$w " + word);
    }
    else
      System.out.println("Could not find method word to say");
  }  // end of methodEntryEvent()



  private void methodExitEvent(MethodExitEvent event)
  /* all code in the method has been executed, and we are about to return;
     play a clip representing the return (or program exit) */
  {  
    Method meth = event.method();
    String className = meth.declaringType().name();

    if (meth.isConstructor())
      System.out.println("exiting " + className + " constructor");
    else
      System.out.println("exiting " + className + "." + meth.name() + "()" );

    // report method return/ program exit
    if (meth.name().equals("main"))   // program is exiting
      saySound("$c keyDrop 430");
    else              // the method is returning
      saySound("$c zap 500");    

    System.out.println();
  }  // end of methodExitEvent()


  // -------------------- class event handling  ---------------


  private void classPrepareEvent(ClassPrepareEvent event)
  /* A new class has been loaded. Print details, parse its file
     to build a collection of notes, and add its methods to the spoken
     words stored in methodWords. */
  {
    ReferenceType ref = event.referenceType();

    List<Field> fields = ref.fields();
    List<Method> methods = ref.methods();
    String className = ref.name();

    String fnm;
    try {
      fnm = ref.sourceName();  // get filename of the class
      appNotes.add(fnm);   // store the musical notes for the code in fnm
    }
    catch (AbsentInformationException e) 
    {  fnm = "??"; }

    System.out.println("loaded class: " + className + " from " + fnm +
                    " - fields=" + fields.size() + ", methods=" + methods.size() );

    methodWords.addMethods(className, methods);   // add the methods to the spoken words
  }  // end of classPrepareEvent()



  // -------------------- thread event handling  ---------------

  private void threadStartEvent(ThreadStartEvent event)
  // a new thread has started running -- switch on single stepping
  {
    ThreadReference thr = event.thread();

    if (thr.name().equals("Signal Dispatcher") || 
        thr.name().equals("DestroyJavaVM") ||
        thr.name().startsWith("AWT-") )     // AWT threads
      return;
    
    if (thr.threadGroup().name().equals("system"))   // ignore system threads
      return;

    // System.out.println(thr.name() + " thread started");

    setStepping(thr);
  } // end of threadStartEvent()



  private void setStepping(ThreadReference thr)
  // start single stepping through the new thread
  {
    EventRequestManager mgr = vm.eventRequestManager();

    StepRequest sr = mgr.createStepRequest(thr, StepRequest.STEP_LINE,
                                                StepRequest.STEP_INTO);
    sr.setSuspendPolicy(EventRequest.SUSPEND_EVENT_THREAD);

    for (int i = 0; i < excludes.length; ++i)
      sr.addClassExclusionFilter(excludes[i]);
    sr.enable();
  }  // end of setStepping()


  // -------------------- step event handling  ---------------


  private void stepEvent(StepEvent event)
  // Play the note for the line that's about to be executed (if there is one)
  { 
   Location loc = event.location();

    try {   // get the line of code
      String fnm = loc.sourceName();  // get filename of code
      int lineNum = loc.lineNumber();
      String note = appNotes.getNote(fnm, lineNum);

      // play note for the line
      if (note != null) {
        // System.out.println(fnm + "." + lineNum + ": play " + note);
        saySound("$n " + note);
      }
    }
    catch (AbsentInformationException e) {}
  }  // end of stepEvent()



  // ---------------------- VM event handling ----------------------------------

  private void vmStartEvent(VMStartEvent event)
  /* Notification of initialization of a target VM. This event is received 
     before the main thread is started and before any application code has 
     been executed. */
  { vmDied = false;
    System.out.println("-- VM Started --"); 
  }


  private void vmDeathEvent(VMDeathEvent event)
  // Notification of VM termination
  { vmDied = true;
    System.out.println("-- The application has exited --");
  }


  private void vmDisconnectEvent(VMDisconnectEvent event)
  /* Notification of disconnection from the VM, either through normal termination 
     or because of an exception/error. */
  { connected = false;
    if (!vmDied)
      System.out.println("-- The application has been disconnected --");
  }



  // ---------------------- sound queue comms ----------------------------------


  private void saySound(String msg)
  // add msg to sound queue so it can be spoken/played
  {
    try {
      sndQueue.put(msg);
    }
    catch(InterruptedException e) {}
  }  // end of saySound()

}  // end of SoundEventsMonitor class
