
// SoundGen.java
// Andrew Davison, April 2009, ad@fivedots.coe.psu.ac.th

/* The SoundGen thread monitors a blocking queue. 
   Periodically (every DELAY ms) it drains the queue of messages,
   simplifies the list, and  converts the remaining messages into
   sounds in different ways.

   Either a message is spoken (using freeTTS), a note is played
   (using Java Sound MIDI), or a WAV audio clip is played.

   Three types of messages arrive on the BlockingQueue:
     $w <word to be said>
     $n <note to be played>
     $c <wav filename (a clip) to be played>  <duration(ms)>
           the file is assumed to be in Sounds/

   The messages are sent from SoundEventMonitor which is tracing an
   executing program using JDI.
*/


import java.util.*;
import java.util.concurrent.BlockingQueue;
 

public class SoundGen extends Thread 
{
  private static final int DELAY = 500;  // time delay between queue examinations

  private static final int CLIP_DURATION = 2000;  // fixed duration for playing a clip

  private static final int NOTE_DURATION = 800;  // fixed duration for playing a note
  private static final int MAX_REPEATS = 3;  // max multiple for playing repeated notes


  private BlockingQueue<String> queue;   // for holding incoming messages
  private boolean progFinished = false;

  // sound processing objects
  private Speaker speaker;
  private NotesPlayer notesPlayer;
  private ClipsPlayer clipsPlayer;


  public SoundGen(BlockingQueue<String> q) 
  {
    queue = q;

    // initialize sound playing tools
    speaker = new Speaker();
    notesPlayer = new NotesPlayer();
    clipsPlayer = new ClipsPlayer();
  }  // end of SoundGen()

 
  public void run()
  // periodically drain and process messages
  {
    ArrayList<String> msgs = new ArrayList<String>();   
             // to store messages removed from the queue
    try {  
      while (!progFinished) {
        msgs.clear();
        queue.drainTo(msgs);    // the draining may not return anything
        combineMsgs(msgs);
        for(String msg : msgs)
          processMsg(msg);
        wait(DELAY);
      }

      // finish off
      System.out.println("thread exiting");
      speaker.closeDown(); 
	  notesPlayer.closeDown();
    }
    catch (Exception e) 
    {  System.out.println(e); }
  }  // end of run()



  private void combineMsgs(ArrayList<String> msgs)
  /* There are potentially lots of ways to simplify the messages. At present
     this code only looks for repeated messages and replaces them with a single
     message followed by a number of repeats.
  */
  {
    // showMessages("drained", msgs);

    int totalRemoved = 0;
    for(int i=0; i < msgs.size(); i++)
      totalRemoved += removeDups(i, msgs);     // the msgs list may get smaller

    // if (totalRemoved > 0)
    //  showMessages("after combinations", msgs);
  }  // end of combineMsgs()



  private void showMessages(String title, ArrayList<String> msgs)
  {
    System.out.println("No. of messages " + title + ": " + msgs.size());
    for (String msg : msgs)
      System.out.print("\"" + msg + "\"  ");
    System.out.println();
  }  // end of showMessages()



  private int removeDups(int i, ArrayList<String> msgs)
  /* starting from i+1, remove any duplicates of the message at
     position i in the list. Change the message to reflect 
     the removals by adding a number to its end. */
  {
    String currMsg = msgs.get(i);
    int pos = i+1;

    int numDups = 0;
    while ((pos+numDups) < msgs.size()) {
      if (!currMsg.equals( msgs.get(pos+numDups) ))
        break;
      numDups++;
    }

    if (numDups > 0) {
      System.out.println("Removing " + numDups + " duplicates" );
      int count = 0;
      while (count != numDups) {
        msgs.remove(pos);
        count++;
      }

      // add number of duplicates to the end of the current message at i
      msgs.set(i, currMsg + (" " + (numDups+1)) );
    }

    return numDups;
  }  // end of removeDups()




  private void processMsg(String msg)
  // generate a sound for a message in three different ways
  {
    System.out.println("--> " + msg);

    if (msg.startsWith("$w "))        // speak a word
      speaker.say( msg.substring(3) );
    else if (msg.startsWith("$n "))   // play a note
      processNote( msg.substring(3) );
    else if (msg.startsWith("$c "))   // play a clip
      processClip( msg.substring(3) );
    else
      System.out.println("Could not understand " + msg);
  }  // end of processMsg()




  private void processNote(String msg)
  // msg format: <note to be played>  [ < number of repeats > ]
  {
    String[] toks = msg.split("\\s+");
    
    int numTimes = 0;
    if (toks.length == 2) {   // there is a number
      try {
        numTimes = Integer.parseInt(toks[1]);
      }
      catch (Exception e)
      { System.out.println("Number is incorrect for " + toks[1]);  }
    }

    if (numTimes != 0) {  // play a note repeatedly, and more quickly
      if (numTimes > MAX_REPEATS)
        numTimes = MAX_REPEATS;
      for (int i=0; i < numTimes; i++)
        notesPlayer.play(toks[0], NOTE_DURATION/2); 
    }
    else                 // play a single note
      notesPlayer.play(toks[0], NOTE_DURATION);            
  }  // end of processNote()




  private void processClip(String msg)
  // msg format: <wav filename (a clip) to be played>  <duration(ms)> [ < number of repeats > ]
  {
    String[] toks = msg.split("\\s+");
    
    if (toks[0].equals("keyDrop"))    
      /* denotes end of main() in the tracing program, which means the 
         end of tracing, and of sound generation */
      progFinished = true;

    // get duration info
    int duration = 0;
    if (toks.length >= 2) {   // there is a duration
      try {
        duration = Integer.parseInt(toks[1]);
      }
      catch (Exception e)
      { System.out.println("Duration is incorrect for " + toks[1]);  }
    }
    if (duration == 0)    // problem with duration, so use a guess
      duration = CLIP_DURATION;


    // get number of times info
    int numTimes = 1;
    if (toks.length == 3) {   // there is a number
      try {
        numTimes = Integer.parseInt(toks[2]);
      }
      catch (Exception e)
      { System.out.println("Repeats is incorrect for " + toks[2]);  }
    }

    // play a clip repeatedly
    for (int i=0; i < numTimes; i++)
      clipsPlayer.play(toks[0], duration);
  }  // end of processClip()



  private void wait(int duration)
  {
    try {
      Thread.sleep(duration); // in ms
    } 
    catch (InterruptedException e) 
    { e.printStackTrace(); }
  }  // end of wait()


}  // end of SoundGen class
