
// Speaker.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, April 2009

/* Initialize a freeTTS "general" speech synthesizer using the voice
   specified in VOICE. freeTTS can be obtained from: 
      http://freetts.sourceforge.net

   If the voice is a MBROLA then mbrola.exe and the 3 US voices must be
   downloaded from the Mbrola site http://tcts.fpms.ac.be/synthesis/mbrola.html
    
   A test message is spoken by calling the say() method, which doesn't
    return until the message is finished.

   The synthesizer is closed down by calling closeDown().

   This code does not use the JSAPI Central class, and so there's
   no need to store its speech.properties file in
   JAVA_HOME\jre\lib

   Usage:
     > compileTTS Speaker.java
     > runTTS Speaker "sentence to be said"
 */


import java.util.*;

import com.sun.speech.freetts.jsapi.FreeTTSEngineCentral;
import javax.speech.*;
import javax.speech.synthesis.*;



public class Speaker
{
  private static final String VOICE = "mbrola_us1";       // "kevin16"
            // mbrola_us1 is a mbrola young US female
            // mbrola_us2 and mbrola_us3 are mbrola young US males
            // kevin16 is freeTTS's medium quality general-purpose US voice

  private Synthesizer synthesizer;


  public Speaker()
  { 
	try {
      System.out.println("General Synthesizer using \"" + VOICE + 
                                          "\" voice being initialized...");

      // specify the required synthesizer properties
	  SynthesizerModeDesc desc = new SynthesizerModeDesc(
                null,          // engine name (don't care)
                "general",     // mode name -- general usage
                Locale.US,     // locale
                null,          // prefer a running synthesizer (don't care)
                null);         // voice (none specified yet)

      FreeTTSEngineCentral central = new FreeTTSEngineCentral();   // so no need for "speech.properties"
      EngineList list = central.createEngineList(desc);
      if(list.size() > 0) {
        EngineCreate creator = (EngineCreate)list.get(0);
        synthesizer = (Synthesizer)creator.createEngine();
      }

	  if (synthesizer == null) {
        System.err.println("Cannot create synthesizer");
		System.exit(1);
	  }

	  synthesizer.allocate();   // allocate the specified resources (may take a while)
	  synthesizer.waitEngineState(Synthesizer.ALLOCATED);

      SynthesizerProperties synProps = synthesizer.getSynthesizerProperties();
      synProps.setVoice( getVoice(VOICE)); 

	  synthesizer.resume();     // get synthesizer ready to speak

      // wait until the synthesizer is ready to speak
	  synthesizer.waitEngineState(Synthesizer.RESUMED);
	  synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);
      System.out.println("Synthesizer ready");
	} 
    catch (Exception e) 
    {  System.out.println(e);
       System.exit(1);  
    }
  }  // end of Speaker()



  private Voice getVoice(String voiceName)
  // get the Voice object associated with voiceName
  {
    // get the properties for this synthesizer engine
    SynthesizerModeDesc desc = 
          (SynthesizerModeDesc) synthesizer.getEngineModeDesc();
    Voice[] voices = desc.getVoices();

    // check if voiceName is a known Voice for this engine
    Voice voice = null;
    for(int i = 0; i < voices.length; i++) {
      if (voices[i].getName().equals(voiceName)) {
        voice = voices[i];
        break;
      }
    }
    if (voice == null) {
      System.err.println("Synthesizer could not find the " + voiceName + " voice");
      System.exit(1);
    }

    return voice;
  } // end of getVoice()


  public void say(String msg)
  // request that msg be spoken, and wait for it to be carried out
  {
    try {
	  synthesizer.speakPlainText(msg, null);  // add msg to the speaking queue

     // wait for the queue to empty (i.e. until the msg has been said)
     synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);
	} 
    catch (Exception e) 
    {  System.out.println(e); }
  }  // end of sayMessage()



  public void closeDown()
  // close down the synthesizer
  {
    try {
      synthesizer.cancel();        // cancel any speaking
	  synthesizer.deallocate();    // free the engine's resources
	} 
    catch (Exception e)
    {  System.out.println(e); }
  }  // end of closeDown()



  // -------------------- test rig ----------------------

  public static void main(String[] args)
  {  
    String msg;
    if (args.length != 1) {
      System.out.println("Usage: runTTS Speaker \"sentence to be said\"");
      msg = "You must tell me what to say";
    }
    else
      msg = args[0];

    Speaker speaker = new Speaker();
    speaker.say(msg);
    speaker.closeDown(); 
  }  // end of main()

} // end of Speaker class
